/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// 3D Renderer: Vertex Input

#include "bstone_r3r_vertex_input.h"

namespace bstone {

R3rVertexInput::R3rVertexInput() noexcept = default;

R3rVertexInput::~R3rVertexInput() = default;

} // namespace bstone
