/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// Universally unique identifier (UUID).
//
// Supported formats:
// - "AAAAAAAABBBBCCCCDDDDEEEEEEEEEEEE"
// - "AAAAAAAA-BBBB-CCCC-DDDD-EEEEEEEEEEEE"
// - "{AAAAAAAA-BBBB-CCCC-DDDD-EEEEEEEEEEEE}"

#include "bstone_uuid.h"
