/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// FIXME Move to the configuration?

#ifndef BSTONE_SYS_LIMITS_SDL2_INCLUDED
#define BSTONE_SYS_LIMITS_SDL2_INCLUDED

namespace bstone {
namespace sys {
namespace limits {

constexpr auto max_textures = 2;
constexpr auto max_display_modes = 256;

} // namespace limits
} // namespace sys
} // namespace bstone

#endif // BSTONE_SYS_LIMITS_SDL2_INCLUDED
