#ifndef BSTONE_SYS_DISPLAY_MODE_INCLUDED
#define BSTONE_SYS_DISPLAY_MODE_INCLUDED

namespace bstone {
namespace sys {

struct DisplayMode
{
	int width;
	int height;
	int refresh_rate;
};

} // namespace sys
} // namespace bstone

#endif // BSTONE_SYS_DISPLAY_MODE_INCLUDED
