/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 1992-2013 Apogee Entertainment, LLC
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: GPL-2.0-or-later
*/


#ifndef BSTONE_JM_IO_INCLUDED
#define BSTONE_JM_IO_INCLUDED


int IO_LoadFile(
	const char* filename,
	void** dst);


#endif // BSTONE_JM_IO_INCLUDED
