/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// FourCC stored as integer in little-endian order (i.e. '0123' => 0x33323130).

#include "bstone_four_cc.h"
