/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2023-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// Renderer's swap interval type.

#ifndef BSTONE_SYS_R3R_SWAP_INTERVAL_TYPE_INCLUDED
#define BSTONE_SYS_R3R_SWAP_INTERVAL_TYPE_INCLUDED

namespace bstone {
namespace sys {

enum class SwapIntervalType
{
	none,
	standard,
	adaptive,
};

} // namespace sys
} // namespace bstone

#endif // BSTONE_SYS_R3R_SWAP_INTERVAL_TYPE_INCLUDED
