/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// A stub for DOSBox OPL2 emulator.

#ifndef BSTONE_DOSBOX_SETUP_INCLUDED
#define BSTONE_DOSBOX_SETUP_INCLUDED

struct Section {};
struct Module_base {};

#endif // BSTONE_DOSBOX_SETUP_INCLUDED
