/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/


//
// Nuked OPL3 wrapper.
//

#ifndef BSTONE_NUKED_OPL3_INCLUDED
#define BSTONE_NUKED_OPL3_INCLUDED

#include "bstone_opl3.h"

namespace bstone
{

Opl3UPtr make_nuked_opl3();

} // bstone

#endif // !BSTONE_NUKED_OPL3_INCLUDED
