#include "bstone_configurations.h"

namespace bstone {

extern const std::intptr_t posix_string_capacity = 256 * 1024;
extern const std::intptr_t win32_wstring_capacity = 256 * 1024;

} // namespace bstone
