/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: MIT
*/

// A stub for DOSBox OPL2 emulator.

#ifndef BSTONE_DOSBOX_HARDWARE_INCLUDED
#define BSTONE_DOSBOX_HARDWARE_INCLUDED

enum OPL_Mode
{
	OPL_none,
	OPL_cms,
	OPL_opl2,
	OPL_dualopl2,
	OPL_opl3
};

#endif // BSTONE_DOSBOX_HARDWARE_INCLUDED
