/*
BStone: Unofficial source port of Blake Stone: Aliens of Gold and Blake Stone: Planet Strike
Copyright (c) 1992-2013 Apogee Entertainment, LLC
Copyright (c) 2013-2024 Boris I. Bendovsky (bibendovsky@hotmail.com) and Contributors
SPDX-License-Identifier: GPL-2.0-or-later
*/


//
// Accelerated video (interface).
//


#include "bstone_video.h"

namespace bstone {

Video::Video() noexcept = default;

Video::~Video() = default;

} // namespace bstone
